/*   This is part of ALE4Net
 *   Application Level Environment for Networking
 *   
 *   Copyright 2004 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */   

/*
 * Neelkanth Natu (neelnatu@yahoo.com) is the author of this file.
 * As long as you keep this blurb in the file, feel free to do whatever.
 */ 
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/un.h>
#include <netinet/in.h>

#include <fcntl.h>
#include <errno.h>
#include <assert.h>

#include "native_syscalls.h"
#include "a4n_server.h"

fd_set rdset;
const int rdsetsize = FD_SETSIZE;

int alpine_errno;
//extern int map_errno(int freebsd_errno);

/* 
 * Read the next request off the 'fd' and return 0 if successfully
 * processed. Non-zero otherwise.
 */
extern int read_request(int fd);

static const char *default_listen_path = "/tmp/a4ns";
static const int tap_minor = 0;

static int listening_fd;
static int do_init(void);
static int accept_new_connection(void);
static void a4ns_main_loop(void);

static char *interface;
static struct in_addr alpine_ip;
static struct in_addr alpine_netmask;
static int scheduler_new_connection(int new_fd);
static const char *listen_addr;

static void signalexit()
{
	unlink(listen_addr);
	_exit(1);
}

static void unlinksocket()
{
	unlink(listen_addr);
}

int
do_init(void)
{
	int rc, sin_len, listen_port;
	//struct sockaddr_in sin;
	struct sockaddr_un sin;
	int on = 1;

	/* Create the socket */
	//listening_fd = native_socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	listening_fd = native_socket(AF_UNIX, SOCK_STREAM, 0);
	if (listening_fd == -1) {
		printf("Error opening listening socket: %s\n", strerror(errno));
		return (-1);
	}

	native_setsockopt(listening_fd, SOL_SOCKET, SO_REUSEADDR, 
			  &on, sizeof(on));

	/* XXX: Make non-blocking ? */

	/* Bind it */
	/*listen_addr = getenv("ALPINE_SERVER_LISTEN_ADDR") ?
		      (const char *)getenv("ALPINE_SERVER_LISTEN_ADDR") :
		      default_listen_addr;
	listen_port = getenv("ALPINE_SERVER_LISTEN_PORT") ?
		      atoi(getenv("ALPINE_SERVER_LISTEN_PORT")) :
		      default_listen_port;
	bzero(&sin, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = inet_addr(listen_addr);
	sin.sin_port = htons(listen_port);*/
	listen_addr = getenv("ALPINE_SERVER_PATH") ?
		(const char *)getenv("ALPINE_SERVER_PATH") :
		default_listen_path;

	bzero(&sin, sizeof(sin));
	sin.sun_family = AF_UNIX;
	strcpy(sin.sun_path, listen_addr);

	rc = native_bind(listening_fd, (struct sockaddr *)&sin, sizeof(sin));
	if (rc != 0) {
		//printf("Error binding listening socket to %s:%d: %s\n",
			//listen_addr, listen_port, strerror(errno));
		printf("Error binding listening socket to %s: %s\n",
			listen_addr, strerror(errno));
		return (-2);
	}
	atexit(unlinksocket);
	signal(SIGTERM, signalexit);
	signal(SIGINT, signalexit);

	/* Listen on it */
	rc = native_listen(listening_fd, 5);
	if (rc != 0) {
		printf("Error listen()ing on socket: %s\n", strerror(errno));
		return(-3);
	}

	bzero(&sin, sizeof(sin));
	sin_len = sizeof(sin);
	rc = native_getsockname(listening_fd, 
				(struct sockaddr *)&sin, &sin_len);
	if (rc != 0) {
		printf("getsockname() error: %s\n", strerror(errno));
		return (-4);
	}

	//printf("Listening on %s:%d!\n", inet_ntoa(sin.sin_addr), 
	//			        ntohs(sin.sin_port));
	printf("Listening on %s!\n", sin.sun_path);

	return (0);
}

int
main (int argc, char *argv[])
{
	FD_ZERO(&rdset);
	if (do_init() != 0) {
		printf("Unable to initialize the alpine server\n");
		exit(-1);
	}

	a4ns_main_loop();
}

static int scheduler_new_connection(int new_fd)
{
	FD_SET(new_fd,&rdset);
// if (new_fd >= rdsetsize) rdsetsize=new_fd+1;
	return new_fd;
}

int
accept_new_connection(void)
{
	//struct sockaddr_in sin;
	struct sockaddr_un sin;
	int new_fd, sin_len;
		
	bzero(&sin, sizeof(sin));
	sin_len = sizeof(sin);
	new_fd = native_accept(listening_fd, (struct sockaddr *)&sin, &sin_len);
	if (new_fd >= 0) {
		bzero(&sin, sizeof(sin));
		sin_len = sizeof(sin);
		(void)native_getpeername(new_fd, 
				         (struct sockaddr *)&sin, &sin_len);
		printf("Received new connection from %s = %d\n",
			sin.sun_path, new_fd);
	}
	return (new_fd);
}

void sigpipehandler(int signo) 
{
	printf("Caught a SIGPIPE signal\n");
}

void
a4ns_main_loop()
{
	int n, new_fd, fd;
	fd_set selectset;
	fd_set xceptset;

	signal (SIGPIPE, sigpipehandler);
	FD_SET(listening_fd, &rdset);
	while (1) {
		selectset=xceptset=rdset;
		n = native_select(rdsetsize, &selectset, NULL, &xceptset, NULL);
		if (n == -1) {
			if (errno != EINTR)
				perror("A4NS select failed"); 
			assert(errno == EINTR || errno == EBADF);
			continue;
		}

		if (FD_ISSET(listening_fd, &xceptset)) {
			printf("Xcept listening_fd on %d\n",listening_fd);
		}
		for (fd=0; fd<rdsetsize && n>0; fd++)
			if FD_ISSET(fd,&xceptset) {
				printf("Xcept on %d\n",fd);
			}
		/* Process a new connection */
		if (FD_ISSET(listening_fd, &selectset)) {
			new_fd = accept_new_connection();
			assert((new_fd >= 0) || (errno == EWOULDBLOCK));
			if (new_fd >= 0) {
				int32_t newpid;
				int nwritten;
				newpid = scheduler_new_connection(new_fd);
				nwritten = write_n_bytes(new_fd, &newpid,
							 sizeof(newpid));
				if (nwritten != sizeof(newpid)) {
					printf("Error writing newpid\n");
					native_close(new_fd);
				}
			} 
			n--;
		}
		for (fd=0; fd<rdsetsize && n>0; fd++)
			if FD_ISSET(fd,&selectset) {
				n--;
				if (read_request(fd) < 0) {
					native_close(fd);
					FD_CLR(fd,&rdset);
					printf("closed connection %d\n", fd);
				}
			}
	}

	return;
}
