/*   This is part of ALE4Net
 *   Application Level Environment for Networking
 *   
 *   server_resp_reply: dialogue with server
 *
 *   Copyright 2004 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */   
/*
 * Neelkanth Natu (neelnatu@yahoo.com) is the author of this file.
 * As long as you keep this blurb in the file, feel free to do whatever.
 */ 
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <assert.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>
#include <stdio.h>

#include "req_resp.h"
#include "fd_mgmt.h"
#include "alpine_support.h"

#define REQ_FIELD_SIZE (sizeof(int32_t))

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif

#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif
#define lwip
#ifdef lwip
#define alpine_accept lwip_accept
#define alpine_bind lwip_bind
#define alpine_connect lwip_connect
#define alpine_listen lwip_listen
#define alpine_recvfrom lwip_recvfrom
#define alpine_read lwip_read
#define alpine_recv lwip_recv
#define alpine_send lwip_send
#define alpine_sendto lwip_sendto
#define alpine_socket lwip_socket
#define alpine_write lwip_write
#define alpine_shutdown lwip_shutdown
#define alpine_getpeername lwip_getpeername
#define alpine_getsockname lwip_getsockname
#define alpine_getsockopt lwip_getsockopt
#define alpine_setsockopt lwip_setsockopt
#define alpine_ioctl lwip_ioctl
#define alpine_fcntl lwip_fcntl
#define alpine_close lwip_close
#define alpine_recvmsg a4n_recvmsg
#define alpine_sendmsg a4n_sendmsg
extern int errno;
extern int lwip_errno;
extern int lwip_close(int fd);
extern int lwip_shutdown(int fd);
#define alpine_internal_errno lwip_errno
#else
extern int alpine_errno;
#define alpine_internal_errno alpine_errno
#endif

int read_request(int fd);

typedef int (*request_func) (int fd, enum request_type rtype);

static int request_func_notsupp(int fd, enum request_type type);
static int request_func_socket(int fd, enum request_type rtype);
static int request_func_close(int fd, enum request_type rtype);
static int request_func_bind(int fd, enum request_type type);
static int request_func_listen(int fd, enum request_type type);
static int request_func_accept(int fd, enum request_type type);
static int request_func_connect(int fd, enum request_type type);
static int request_func_read(int fd, enum request_type type);
static int request_func_write(int fd, enum request_type type);
static int request_func_getname(int fd, enum request_type type);
static int request_func_recvfrom(int fd, enum request_type type);
static int request_func_recvmsg(int fd, enum request_type rtype);
static int request_func_sendto(int fd, enum request_type type);
static int request_func_sendmsg(int fd, enum request_type rtype);
static int request_func_shutdown(int fd, enum request_type type);
static int request_func_setsockopt(int fd, enum request_type type);
static int request_func_getsockopt(int fd, enum request_type type);
static int request_func_fcntl(int fd, enum request_type type);
static int request_func_select(int fd, enum request_type type);
static int request_func_fork(int fd, enum request_type type);
static int request_func_dup(int fd, enum request_type type);

static int request_func_ioctl(int fd, enum request_type rtype);

static request_func request_dispatch_table[1 + REQ_LAST] = {
	NULL,
	request_func_socket,
	request_func_close,
	request_func_bind,
	request_func_listen,
	request_func_accept,
	request_func_connect,
	request_func_read,
	request_func_write,
	request_func_getname,	/* REQ_GETSOCKNAME */
	request_func_getname,	/* REQ_GETPEERNAME */
	request_func_recvfrom,	/* REQ_RECV */
	request_func_recvfrom,	/* REQ_RECVFROM */
	request_func_recvmsg,
	request_func_sendto,	/* REQ_SEND */
	request_func_sendto, 	/* REQ_SENDTO */
	request_func_sendmsg,
	request_func_shutdown,
	request_func_setsockopt,
	request_func_getsockopt,
	request_func_fcntl,
	request_func_select,
	request_func_select,
	request_func_fork,
	request_func_dup,

	request_func_ioctl,
};

#define fdck(cfd,fd,call) (fdmgmt_check((cfd),(fd)) ? (call) : ((alpine_internal_errno = EACCES), -1))
		
int 
request_func_notsupp(int fd, enum request_type type)
{
	printf("Unsupported system call: %s\n",
		request_type_to_string[type]);
	assert(0);
}

int
request_func_socket(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct socket_req socket_req;
	struct socket_resp socket_resp;

	assert(fd >= 0);
	assert(rtype == REQ_SOCKET);

	nread = read_n_bytes(fd, 
			     (char *)&socket_req + REQ_FIELD_SIZE,
			     sizeof(socket_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(socket_req) - REQ_FIELD_SIZE) {
		printf("Error reading socket_req\n");
		retval = -1;
		goto end;
	}

	socket_resp.resp_type = RESP_SOCKET;
	socket_resp.retval = alpine_socket(socket_req.domain,
					   socket_req.type,
					   socket_req.protocol);
	socket_resp.alpine_errno = alpine_internal_errno;
	fdmgmt_open(fd, socket_resp.retval);

	nwritten = write_n_bytes(fd, (char *)&socket_resp, sizeof(socket_resp));
	if (nwritten != sizeof(socket_resp)) {
		printf("Error writing socket_resp\n");
		retval = -2;
		goto end;
	}

end:
	return (retval);
}

int
request_func_close(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct close_req close_req;
	struct close_resp close_resp;

	assert(fd >= 0);
	assert(rtype == REQ_CLOSE);

	nread = read_n_bytes(fd, 
			     (char *)&close_req + REQ_FIELD_SIZE,
			     sizeof(close_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(close_req) - REQ_FIELD_SIZE) {
		printf("Error reading close_req\n");
		retval = -1;
		goto end;
	}

	close_resp.resp_type = RESP_CLOSE;
	//close_resp.retval = alpine_close(close_req.fd);
	close_resp.retval = fdck(fd, close_req.fd, fdmgmt_close(fd, close_req.fd, alpine_close));
	close_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, (char *)&close_resp, sizeof(close_resp));
	if (nwritten != sizeof(close_resp)) {
		printf("Error writing close_resp\n");
		retval = -2;
		goto end;
	}

end:
	return (retval);
}

int
request_func_shutdown(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct shutdown_req shutdown_req;
	struct shutdown_resp shutdown_resp;

	assert(fd >= 0);
	assert(rtype == REQ_SHUTDOWN);

	nread = read_n_bytes(fd, 
			     (char *)&shutdown_req + REQ_FIELD_SIZE,
			     sizeof(shutdown_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(shutdown_req) - REQ_FIELD_SIZE) {
		printf("Error reading shutdown_req\n");
		retval = -1;
		goto end;
	}

	shutdown_resp.resp_type = RESP_SHUTDOWN;
	//shutdown_resp.retval=alpine_shutdown(shutdown_req.fd,shutdown_req.how);
	shutdown_resp.retval=fdck(fd, shutdown_req.fd, fdmgmt_close(fd, shutdown_req.fd,alpine_shutdown));
	shutdown_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, 
			         (const char *)&shutdown_resp, 
				 sizeof(shutdown_resp));
	if (nwritten != sizeof(shutdown_resp)) {
		printf("Error writing shutdown_resp\n");
		retval = -2;
		goto end;
	}

end:
	return (retval);
}

int
request_func_bind(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct bind_req bind_req;
	struct bind_resp bind_resp;
	char *myaddr = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_BIND);

	nread = read_n_bytes(fd, 
			     (char *)&bind_req + REQ_FIELD_SIZE,
			     sizeof(bind_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(bind_req) - REQ_FIELD_SIZE) {
		printf("Error reading bind_req\n");
		retval = -1;
		goto end;
	}

	myaddr = (char *)alpine_malloc(MAX(bind_req.addrlen, sizeof(struct sockaddr)));
	assert(myaddr != NULL);

	nread = read_n_bytes(fd, myaddr, bind_req.addrlen);
	if (nread != bind_req.addrlen) {
		printf("Error reading bind_req myaddr\n");
		retval = -2;
		goto end;
	}

	bind_resp.resp_type = RESP_BIND;
	bind_resp.retval = fdck(fd, bind_req.fd, alpine_bind(bind_req.fd, myaddr, bind_req.addrlen));
	bind_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, (char *)&bind_resp, sizeof(bind_resp));
	if (nwritten != sizeof(bind_resp)) {
		printf("Error writing bind_resp\n");
		retval = -3;
		goto end;
	}

end:
	if (myaddr != NULL) {
		alpine_free(myaddr);
	}
	return (retval);
}

int
request_func_connect(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct connect_req connect_req;
	struct connect_resp connect_resp;
	char *addr = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_CONNECT);

	nread = read_n_bytes(fd, 
			     (char *)&connect_req + REQ_FIELD_SIZE,
			     sizeof(connect_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(connect_req) - REQ_FIELD_SIZE) {
		printf("Error reading connect_req\n");
		retval = -1;
		goto end;
	}

	addr = (char *)alpine_malloc(MAX(connect_req.addrlen, sizeof(struct sockaddr)));
	assert(addr != NULL);

	nread = read_n_bytes(fd, addr, connect_req.addrlen);
	if (nread != connect_req.addrlen) {
		printf("Error reading connect_req addr\n");
		retval = -2;
		goto end;
	}

	connect_resp.resp_type = RESP_CONNECT;
	connect_resp.retval = fdck (fd, connect_req.fd, alpine_connect(connect_req.fd, 
					     addr, connect_req.addrlen));
	connect_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, 
				 (char *)&connect_resp, sizeof(connect_resp));
	if (nwritten != sizeof(connect_resp)) {
		printf("Error writing connect_resp\n");
		retval = -3;
		goto end;
	}

end:
	if (addr != NULL) {
		alpine_free(addr);
	}
	return (retval);
}

int
request_func_read(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct read_req read_req;
	struct read_resp read_resp;
	char *data = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_READ);

	nread = read_n_bytes(fd, 
			     (char *)&read_req + REQ_FIELD_SIZE,
			     sizeof(read_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(read_req) - REQ_FIELD_SIZE) {
		printf("Error reading read_req\n");
		retval = -1;
		goto end;
	}

	if (read_req.count > 0) {
		data = (char *)alpine_malloc(read_req.count);
		assert(data != NULL);
	}

	read_resp.resp_type = RESP_READ;
	read_resp.retval = fdck (fd, read_req.fd, alpine_read(read_req.fd, data, read_req.count));
	read_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, 
				 (char *)&read_resp, sizeof(read_resp));
	if (nwritten != sizeof(read_resp)) {
		printf("Error writing read_resp\n");
		retval = -3;
		goto end;
	}

	if (read_resp.retval > 0) {
		assert(data != NULL);
		nwritten = write_n_bytes(fd, data, read_resp.retval);
		if (nwritten != read_resp.retval) {
			printf("Error writing read_resp data\n");
			retval = -4;
			goto end;
		}
	}

end:
	if (data != NULL) {
		alpine_free(data);
	}
	return (retval);
}

int
request_func_write(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct write_req write_req;
	struct write_resp write_resp;
	char *data = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_WRITE);

	nread = read_n_bytes(fd, 
			     (char *)&write_req + REQ_FIELD_SIZE,
			     sizeof(write_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(write_req) - REQ_FIELD_SIZE) {
		printf("Error reading write_req\n");
		retval = -1;
		goto end;
	}

	if (write_req.count > 0) {
		data = (char *)alpine_malloc(write_req.count);
		assert(data != NULL);
		nread = read_n_bytes(fd, data, write_req.count);
		if (nread != write_req.count) {
			printf("Error reading write_req data\n");
			retval = -2;
			goto end;
		}
	}

	write_resp.resp_type = RESP_WRITE;
	write_resp.retval = fdck (fd, write_req.fd, alpine_write(write_req.fd, data, write_req.count));
	write_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, 
				 (char *)&write_resp, sizeof(write_resp));
	if (nwritten != sizeof(write_resp)) {
		printf("Error writing write_resp\n");
		retval = -3;
		goto end;
	}

end:
	if (data != NULL) {
		alpine_free(data);
	}
	return (retval);
}

int
request_func_setsockopt(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct setsockopt_req setsockopt_req;
	struct setsockopt_resp setsockopt_resp;
	char *optval = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_SETSOCKOPT);

	nread = read_n_bytes(fd, 
			     (char *)&setsockopt_req + REQ_FIELD_SIZE,
			     sizeof(setsockopt_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(setsockopt_req) - REQ_FIELD_SIZE) {
		printf("Error reading setsockopt_req\n");
		retval = -1;
		goto end;
	}
	assert(setsockopt_req.optlen >= 0);

	if (setsockopt_req.optlen > 0) {
		optval = (char *)alpine_malloc(setsockopt_req.optlen);
		assert(optval != NULL);
		nread = read_n_bytes(fd, optval, setsockopt_req.optlen);
		if (nread != setsockopt_req.optlen) {
			printf("Error reading setsockopt_req optval\n");
			retval = -2;
			goto end;
		}
	}

	setsockopt_resp.resp_type = RESP_SETSOCKOPT;
	setsockopt_resp.retval = fdck (fd, setsockopt_req.fd, alpine_setsockopt(setsockopt_req.fd, 
					setsockopt_req.level,
					setsockopt_req.optname,
					optval, setsockopt_req.optlen));
	setsockopt_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, 
			 (char *)&setsockopt_resp, sizeof(setsockopt_resp));
	if (nwritten != sizeof(setsockopt_resp)) {
		printf("Error writing setsockopt_resp\n");
		retval = -3;
		goto end;
	}

end:
	if (optval != NULL) {
		alpine_free(optval);
	}
	return (retval);
}

int
request_func_getsockopt(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct getsockopt_req getsockopt_req;
	struct getsockopt_resp getsockopt_resp;
	char *optval = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_GETSOCKOPT);

	nread = read_n_bytes(fd, 
			     (char *)&getsockopt_req + REQ_FIELD_SIZE,
			     sizeof(getsockopt_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(getsockopt_req) - REQ_FIELD_SIZE) {
		printf("Error reading getsockopt_req\n");
		retval = -1;
		goto end;
	}
	assert(getsockopt_req.optlen >= 0);

	if (getsockopt_req.optlen > 0) {
		optval = (char *)alpine_malloc(getsockopt_req.optlen);
		assert(optval != NULL);
		nread = read_n_bytes(fd, optval, getsockopt_req.optlen);
		if (nread != getsockopt_req.optlen) {
			printf("Error reading getsockopt_req optval\n");
			retval = -2;
			goto end;
		}
	}

	getsockopt_resp.resp_type = RESP_GETSOCKOPT;
	getsockopt_resp.optlen = getsockopt_req.optlen;
	getsockopt_resp.retval = fdck (fd, getsockopt_req.fd, alpine_getsockopt(getsockopt_req.fd, 
					getsockopt_req.level,
					getsockopt_req.optname,
					optval, &getsockopt_resp.optlen));
	getsockopt_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, 
			 (char *)&getsockopt_resp, sizeof(getsockopt_resp));
	if (nwritten != sizeof(getsockopt_resp)) {
		printf("Error writing getsockopt_resp\n");
		retval = -3;
		goto end;
	}

	assert(getsockopt_resp.optlen >= 0);
	if (getsockopt_resp.optlen > 0) {
		assert(optval != NULL);
		nwritten = write_n_bytes(fd, optval, getsockopt_resp.optlen);
		if (nwritten != getsockopt_resp.optlen) {
			printf("Error writing getsockopt_resp optval\n");
			retval = -4;
			goto end;
		}
	}
end:
	if (optval != NULL) {
		alpine_free(optval);
	}
	return (retval);
}

int
request_func_accept(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct accept_req accept_req;
	struct accept_resp accept_resp;
	char *addr = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_ACCEPT);

	nread = read_n_bytes(fd, 
			     (char *)&accept_req + REQ_FIELD_SIZE,
			     sizeof(accept_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(accept_req) - REQ_FIELD_SIZE) {
		printf("Error reading accept_req\n");
		retval = -1;
		goto end;
	}

	assert(accept_req.addrlen >= 0);
	if (accept_req.addrlen) {
		addr = (char *)alpine_malloc(accept_req.addrlen);
		assert(addr != NULL);

		nread = read_n_bytes(fd, addr, accept_req.addrlen);
		if (nread != accept_req.addrlen) {
			printf("Error reading accept_req addr\n");
			retval = -2;
			goto end;
		}
	}

	accept_resp.resp_type = RESP_ACCEPT;
	accept_resp.addrlen = accept_req.addrlen;
	accept_resp.retval = fdck (fd, accept_req.fd, alpine_accept(accept_req.fd, 
			   addr, 
			   accept_req.addrlen ? &accept_resp.addrlen: NULL));
	accept_resp.alpine_errno = alpine_internal_errno;
	fdmgmt_open(fd,accept_resp.retval);

	nwritten = write_n_bytes(fd, (char *)&accept_resp, sizeof(accept_resp));
	if (nwritten != sizeof(accept_resp)) {
		printf("Error writing accept_resp\n");
		retval = -3;
		goto end;
	}
	if (accept_resp.addrlen) {
		nwritten = write_n_bytes(fd, addr, accept_resp.addrlen);
		if (nwritten != accept_resp.addrlen) {
			printf("Error writing accept_resp addr\n");
			retval = -4;
			goto end;
		}
	}

end:
	if (addr != NULL) {
		alpine_free(addr);
	}
	return (retval);
}

int
request_func_recvmsg(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct recvmsg_req recvmsg_req;
	struct recvmsg_resp recvmsg_resp;
	char *name = NULL, *control = NULL, *buf = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_RECVMSG);

	nread = read_n_bytes(fd, 
			     (char *)&recvmsg_req + REQ_FIELD_SIZE,
			     sizeof(recvmsg_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(recvmsg_req) - REQ_FIELD_SIZE) {
		printf("Error reading recvmsg_req\n");
		retval = -1;
		goto end;
	}

	if (recvmsg_req.namelen > 0) {
		name = (char *)alpine_malloc(recvmsg_req.namelen);
		assert(name != NULL);

	}

	if (recvmsg_req.controllen > 0) {
		control = (char *)alpine_malloc(recvmsg_req.controllen);
		assert(control != NULL);
	}

	if (recvmsg_req.buflen > 0) {
		buf = (char *)alpine_malloc(recvmsg_req.buflen);
		assert(buf != NULL);
	}

	recvmsg_resp.resp_type = RESP_RECVMSG;
	recvmsg_resp.namelen = recvmsg_req.namelen;
	recvmsg_resp.controllen = recvmsg_req.controllen;
	recvmsg_resp.retval = fdck (fd, recvmsg_req.fd, alpine_recvmsg(recvmsg_req.fd,
					   name, &recvmsg_resp.namelen,
					   control, &recvmsg_resp.controllen,
					   buf, recvmsg_req.buflen, 
					   &recvmsg_req.flags));
	recvmsg_resp.alpine_errno = alpine_internal_errno;

	/* Write the response header */
	nwritten=write_n_bytes(fd,(char *)&recvmsg_resp,sizeof(recvmsg_resp));
	if (nwritten != sizeof(recvmsg_resp)) {
		printf("Error writing recvmsg_resp\n");
		retval = -3;
		goto end;
	}
	/* And the 'name' if appropriate */
	if (recvmsg_resp.namelen > 0) {
		nwritten = write_n_bytes(fd, name, recvmsg_resp.namelen);
		if (nwritten != recvmsg_resp.namelen) {
			printf("Error writing recvmsg_resp name\n");
			retval = -4;
			goto end;
		}
	}
	/* And the 'control' if appropriate */
	if (recvmsg_resp.controllen > 0) {
		nwritten = write_n_bytes(fd, control, recvmsg_resp.controllen);
		if (nwritten != recvmsg_resp.controllen) {
			printf("Error writing recvmsg_resp control\n");
			retval = -4;
			goto end;
		}
	}
	/* And the data if appropriate */
	if (recvmsg_resp.retval > 0) {
		nwritten = write_n_bytes(fd, buf, recvmsg_resp.retval);
		if (nwritten != recvmsg_resp.retval) {
			printf("Error writing recvmsg_resp buf\n");
			retval = -4;
			goto end;
		}
	}

end:
	if (name != NULL) {
		alpine_free(name);
	}
	if (control != NULL) {
		alpine_free(control);
	}
	if (buf != NULL) {
		alpine_free(buf);
	}
	return (retval);
}

int
request_func_recvfrom(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct recvfrom_req recvfrom_req;
	struct recvfrom_resp recvfrom_resp;
	char *buf = NULL;
	char *from = NULL;

	assert(fd >= 0);
	assert((rtype == REQ_RECVFROM) || (rtype == REQ_RECV));

	nread = read_n_bytes(fd, 
			     (char *)&recvfrom_req + REQ_FIELD_SIZE,
			     sizeof(recvfrom_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(recvfrom_req) - REQ_FIELD_SIZE) {
		printf("Error reading recvfrom_req\n");
		retval = -1;
		goto end;
	}

	assert(recvfrom_req.len >= 0);
	if (recvfrom_req.len) {
		buf = (char *)alpine_malloc(recvfrom_req.len);
		assert(buf != NULL);

	}

	assert(recvfrom_req.fromlen >= 0);
	if (recvfrom_req.fromlen) {
		from = (char *)alpine_malloc(recvfrom_req.fromlen);
		assert(from != NULL);
	}

	if (rtype == REQ_RECV) {
		recvfrom_resp.resp_type = RESP_RECV;
	} else {
		recvfrom_resp.resp_type = RESP_RECVFROM;
	}
	recvfrom_resp.fromlen = recvfrom_req.fromlen;
	recvfrom_resp.retval =  fdck (fd, recvfrom_req.fd, alpine_recvfrom(recvfrom_req.fd,
			   buf, recvfrom_req.len, recvfrom_req.flags,
			   from, 
			   recvfrom_req.fromlen ? &recvfrom_resp.fromlen: NULL));
	recvfrom_resp.alpine_errno = alpine_internal_errno;

	/* Write the response header */
	nwritten=write_n_bytes(fd,(char *)&recvfrom_resp,sizeof(recvfrom_resp));
	if (nwritten != sizeof(recvfrom_resp)) {
		printf("Error writing recvfrom_resp\n");
		retval = -3;
		goto end;
	}
	/* And the data if appropriate */
	if (recvfrom_resp.retval > 0) {
		nwritten = write_n_bytes(fd, buf, recvfrom_resp.retval);
		if (nwritten != recvfrom_resp.retval) {
			printf("Error writing recvfrom_resp buf\n");
			retval = -4;
			goto end;
		}
	}
	/* And the 'from' sockaddr if appropriate */
	if (recvfrom_resp.fromlen) {
		nwritten = write_n_bytes(fd, from, recvfrom_resp.fromlen);
		if (nwritten != recvfrom_resp.fromlen) {
			printf("Error writing recvfrom_resp from\n");
			retval = -5;
			goto end;
		}
	}

end:
	if (from != NULL) {
		alpine_free(from);
	}
	if (buf != NULL) {
		alpine_free(buf);
	}
	return (retval);
}

int
request_func_sendmsg(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct sendmsg_req sendmsg_req;
	struct sendmsg_resp sendmsg_resp;
	char *name = NULL, *control = NULL, *buf = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_SENDMSG);

	/* Read the sendmsg header */
	nread = read_n_bytes(fd, 
			     (char *)&sendmsg_req + REQ_FIELD_SIZE,
			     sizeof(sendmsg_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(sendmsg_req) - REQ_FIELD_SIZE) {
		printf("Error reading sendmsg_req\n");
		retval = -1;
		goto end;
	}
	/* And msg_name if appropriate */
	if (sendmsg_req.namelen > 0) {
		name = (char *)alpine_malloc(sendmsg_req.namelen);
		assert(name != NULL);
		nread = read_n_bytes(fd, name, sendmsg_req.namelen);
		if (nread != sendmsg_req.namelen) {
			printf("Error reading sendmsg_req.name\n");
			retval = -2;
			goto end;
		}
	}

	/* And msg_control if appropriate */
	if (sendmsg_req.controllen > 0) {
		control = (char *)alpine_malloc(sendmsg_req.controllen);
		assert(control != NULL);
		nread = read_n_bytes(fd, control, sendmsg_req.controllen);
		if (nread != sendmsg_req.controllen) {
			printf("Error reading sendmsg_req.control\n");
			retval = -3;
			goto end;
		}			
	}
	/* And the msg_data if appropriate */
	if (sendmsg_req.buflen > 0) {
		buf = (char *)alpine_malloc(sendmsg_req.buflen);
		assert(buf != NULL);
		nread = read_n_bytes(fd, buf, sendmsg_req.buflen);
		if (nread != sendmsg_req.buflen) {
			printf("Error reading sendmsg_req.buf\n");
			retval = -4;
			goto end;
		}
	}

	bzero(&sendmsg_resp, sizeof(sendmsg_resp));
	sendmsg_resp.resp_type = RESP_SENDMSG;
	sendmsg_resp.retval =  fdck (fd, sendmsg_req.fd, alpine_sendmsg(sendmsg_req.fd,
					     name, sendmsg_req.namelen,
					     control, sendmsg_req.controllen,
					     buf, sendmsg_req.buflen, 
					     sendmsg_req.flags));
	sendmsg_resp.alpine_errno = alpine_internal_errno;

	/* Write the response header */
	nwritten=write_n_bytes(fd,(char *)&sendmsg_resp,sizeof(sendmsg_resp));
	if (nwritten != sizeof(sendmsg_resp)) {
		printf("Error writing sendmsg_resp\n");
		retval = -5;
		goto end;
	}

end:
	if (name != NULL) {
		alpine_free(name);
	}
	if (control != NULL) {
		alpine_free(control);
	}
	if (buf != NULL) {
		alpine_free(buf);
	}
	return (retval);
}

int
request_func_sendto(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct sendto_req sendto_req;
	struct sendto_resp sendto_resp;
	char *buf = NULL;
	char *to = NULL;

	assert(fd >= 0);
	assert((rtype == REQ_SENDTO) || (rtype == REQ_SEND));

	nread = read_n_bytes(fd, 
			     (char *)&sendto_req + REQ_FIELD_SIZE,
			     sizeof(sendto_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(sendto_req) - REQ_FIELD_SIZE) {
		printf("Error reading sendto_req\n");
		retval = -1;
		goto end;
	}

	assert(sendto_req.len >= 0);
	if (sendto_req.len) {
		buf = (char *)alpine_malloc(sendto_req.len);
		assert(buf != NULL);
		nread = read_n_bytes(fd, buf, sendto_req.len);
		if (nread != sendto_req.len) {
			printf("Error reading sendto_req buf\n");
			retval = -2;
			goto end;
		}
	}

	assert(sendto_req.tolen >= 0);
	if (sendto_req.tolen) {
		to = (char *)alpine_malloc(sendto_req.tolen);
		assert(to != NULL);
		nread = read_n_bytes(fd, to, sendto_req.tolen);
		if (nread != sendto_req.tolen) {
			printf("Error reading sendto_req to\n");
			retval = -3;
			goto end;
		}
	}

	if (rtype == REQ_SEND) {
		sendto_resp.resp_type = RESP_SEND;
	} else {
		sendto_resp.resp_type = RESP_SENDTO;
	}
	sendto_resp.retval =  fdck (fd, sendto_req.fd, alpine_sendto(sendto_req.fd,
				   buf, sendto_req.len, sendto_req.flags,
				   to, sendto_req.tolen));
	sendto_resp.alpine_errno = alpine_internal_errno;

	/* Write the response header */
	nwritten=write_n_bytes(fd,(char *)&sendto_resp,sizeof(sendto_resp));
	if (nwritten != sizeof(sendto_resp)) {
		printf("Error writing sendto_resp\n");
		retval = -4;
		goto end;
	}

end:
	if (to != NULL) {
		alpine_free(to);
	}
	if (buf != NULL) {
		alpine_free(buf);
	}
	return (retval);
}

int
request_func_getname(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct getname_req getname_req;
	struct getname_resp getname_resp;
	char *name = NULL;

	assert(fd >= 0);
	assert((rtype == REQ_GETSOCKNAME) || (rtype == REQ_GETPEERNAME));

	nread = read_n_bytes(fd, 
			     (char *)&getname_req + REQ_FIELD_SIZE,
			     sizeof(getname_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(getname_req) - REQ_FIELD_SIZE) {
		printf("Error reading getname_req\n");
		retval = -1;
		goto end;
	}

	assert(getname_req.namelen >= 0);
	if (getname_req.namelen) {
		name = (char *)alpine_malloc(getname_req.namelen);
		assert(name != NULL);
	}

	getname_resp.resp_type = (rtype == REQ_GETSOCKNAME) ? 
		                 RESP_GETSOCKNAME : RESP_GETPEERNAME;
	getname_resp.namelen = getname_req.namelen;
	if (rtype == REQ_GETSOCKNAME) {
		getname_resp.retval = fdck (fd, getname_req.fd, alpine_getsockname(getname_req.fd, 
							 name, 
			                                 getname_req.namelen ? 
							 &getname_resp.namelen:
							 NULL));
	} else {
		getname_resp.retval = alpine_getpeername(getname_req.fd, 
							 name, 
			                                 getname_req.namelen ? 
							 &getname_resp.namelen:
							 NULL);
	}
	getname_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd,(char *)&getname_resp,sizeof(getname_resp));
	if (nwritten != sizeof(getname_resp)) {
		printf("Error writing getname_resp\n");
		retval = -3;
		goto end;
	}
	if (getname_resp.namelen) {
		nwritten = write_n_bytes(fd, name, getname_resp.namelen);
		if (nwritten != getname_resp.namelen) {
			printf("Error writing getname_resp name\n");
			retval = -4;
			goto end;
		}
	}

end:
	if (name != NULL) {
		alpine_free(name);
	}
	return (retval);
}

int
request_func_listen(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct listen_req listen_req;
	struct listen_resp listen_resp;

	assert(fd >= 0);
	assert(rtype == REQ_LISTEN);

	nread = read_n_bytes(fd, 
			     (char *)&listen_req + REQ_FIELD_SIZE,
			     sizeof(listen_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(listen_req) - REQ_FIELD_SIZE) {
		printf("Error reading listen_req\n");
		retval = -1;
		goto end;
	}

	listen_resp.resp_type = RESP_LISTEN;
	listen_resp.retval = fdck (fd, listen_req.fd, alpine_listen(listen_req.fd, listen_req.backlog));
	listen_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, (char *)&listen_resp, sizeof(listen_resp));
	if (nwritten != sizeof(listen_resp)) {
		printf("Error writing listen_resp\n");
		retval = -3;
		goto end;
	}

end:
	return (retval);
}

int 
request_func_ioctl(int fd, enum request_type rtype)
{
	struct ioctl_req ioctl_req;
	struct ioctl_resp ioctl_resp;
	caddr_t data = NULL;
	int retval = 0, nread, nwritten;
	struct ifconf *ifc = NULL;

	assert(fd >= 0);
	assert(rtype == REQ_IOCTL);

	/* Read the ioctl request header */
	nread = read_n_bytes(fd,
			     (char *)&ioctl_req + REQ_FIELD_SIZE,
			     sizeof(ioctl_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(ioctl_req) - REQ_FIELD_SIZE) {
		printf("Error reading ioctl_req\n");
		retval = -1;
		goto end;
	}

	/* Read the data (if any) */
	if (ioctl_req.data_length > 0) {
		data = (caddr_t)alpine_malloc(ioctl_req.data_length);
		nread = read_n_bytes(fd, data, ioctl_req.data_length);
		if (nread != ioctl_req.data_length) {
			printf("Error reading ioctl_req data\n");
			retval = -2;
			goto end;
		}
	}

	/* XXX */
	switch(ioctl_req.command) {
		case SIOCGIFCONF:
			ifc = (struct ifconf *)data;
			if (ifc->ifc_len > 0) {
				ifc->ifc_buf = (caddr_t)
						alpine_malloc(ifc->ifc_len);
				assert(ifc->ifc_buf != NULL);	
			} else {
				ifc->ifc_buf = NULL;
			}
			break;

		default:
			break;
	}

	/* Do the ioctl */
	ioctl_resp.resp_type = RESP_IOCTL;
	ioctl_resp.retval =  fdck (fd, ioctl_req.fd, alpine_ioctl(ioctl_req.fd, ioctl_req.command, data));
	ioctl_resp.alpine_errno = alpine_internal_errno;

	/* Write out the response header */
	nwritten = write_n_bytes(fd, (char *)&ioctl_resp, sizeof(ioctl_resp));
	if (nwritten != sizeof(ioctl_resp)) {
		printf("Error writing ioctl response header\n");
		retval = -4;
		goto end;
	}
	/* 
	 * Write out response data (if any). The client implicitly expects
	 * the same length of data back.
	 */
	if (data != NULL) {
		nwritten = write_n_bytes(fd, data, ioctl_req.data_length);
		if (nwritten != ioctl_req.data_length) {
			printf("Error writing ioctl resp data\n");
			retval = -5;
			goto end;
		}
	}

	/* XXX */
	switch (ioctl_req.command) {
		case SIOCGIFCONF:
			ifc = (struct ifconf *)data;
			if (ifc->ifc_len > 0) {
				assert(ifc->ifc_buf != NULL);
				nwritten = write_n_bytes(fd, ifc->ifc_buf,
							 ifc->ifc_len);
				if (nwritten != ifc->ifc_len) {
					printf("Error writing ioctl ifc_buf\n");
					retval = -6;
					goto end;
				}
			}
			if (ifc->ifc_buf != NULL) {
				alpine_free(ifc->ifc_buf);
			}
			break;
		
		default:
			break;
	}

	/* SUCCESS */
	retval = 0;

end:
	if (data != NULL) {
		alpine_free(data);
	}
	return (retval);
}

int
request_func_fcntl(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct fcntl_req fcntl_req;
	struct fcntl_resp fcntl_resp;

	assert(fd >= 0);
	assert(rtype == REQ_FCNTL);

	nread = read_n_bytes(fd, 
			     (char *)&fcntl_req + REQ_FIELD_SIZE,
			     sizeof(fcntl_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(fcntl_req) - REQ_FIELD_SIZE) {
		printf("Error reading fcntl_req\n");
		retval = -1;
		goto end;
	}

	fcntl_resp.resp_type = RESP_FCNTL;
	fcntl_resp.retval = fdck (fd, fcntl_req.fd,  alpine_fcntl(fcntl_req.fd, fcntl_req.cmd,
					 fcntl_req.arg));
	fcntl_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, (char *)&fcntl_resp, sizeof(fcntl_resp));
	if (nwritten != sizeof(fcntl_resp)) {
		printf("Error writing fcntl_resp\n");
		retval = -2;
		goto end;
	}

end:
	return (retval);
}

int
read_request(int fd)
{
	int n = -2;
	enum request_type req = REQ_NONE;

	/* Read the first 4 bytes of the message */
	req = read_req_type(fd);
	if ((req > REQ_NONE) && (req <= REQ_LAST)) {
		n = (*request_dispatch_table[req])(fd, req);
	} else {
		n = -1;
	}

	if (n<0) {
		fdmgmt_controlclose(fd,alpine_close);
	}

	return (n);
}

static void select_callback(int fd, int nfd, int retval,  fd_set *in, fd_set *out, fd_set *ex);
int
request_func_select(int fd, enum request_type rtype)
{
	int retval = 0, nread;
	struct select_req select_req;
	fd_mask *in = NULL, *out = NULL, *ex = NULL;

	assert(fd >= 0);
	assert((rtype == REQ_SELECT) || (rtype == REQ_SELECT_STOP));

	if (rtype == REQ_SELECT_STOP) {
		lwip_select_cancel(fd);
#if 0
		printf("Race condition between select_resp and select_stop\n");
#endif
		return (0);
	}

	nread = read_n_bytes(fd, 
			     (char *)&select_req + REQ_FIELD_SIZE,
			     sizeof(select_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(select_req) - REQ_FIELD_SIZE) {
		printf("Error reading select_req\n");
		retval = -1;
		return(retval);
	}
	assert(select_req.nfd > 0);

#define read_fd_mask(name, nfds) \
	do { \
		int __len = howmany(nfds, NFDBITS)*sizeof(fd_mask); \
		name = (fd_mask *)alpine_malloc(/*__len*/sizeof(fd_set));	\
		assert(name != NULL); \
		FD_ZERO((fd_set *)name); \
		nread = read_n_bytes(fd, (char *)name, __len); \
		if (nread != __len) { \
			printf("Error reading select_req %s fd_mask\n", \
				#name); \
			retval = -2; \
			goto end; \
		} \
	} while (0)

	read_fd_mask(in, select_req.nfd);
	read_fd_mask(out, select_req.nfd);
	read_fd_mask(ex, select_req.nfd);
#undef read_fd_mask

	lwip_select_check(select_callback, fd, select_req.nfd, (fd_set *)in, (fd_set *)out, (fd_set *)ex);
	return(0);
end:
	if (in != NULL) {
		alpine_free(in);
	}
	if (out != NULL) {
		alpine_free(out);
	}
	if (ex != NULL) {
		alpine_free(ex);
	}
	return(0);
}

static void select_callback(int fd, int nfd, int retval,  fd_set *in, fd_set *out, fd_set *ex)
{
	int nwritten;
	struct select_resp select_resp;
	select_resp.resp_type = RESP_SELECT;
	select_resp.retval = retval; 
	select_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, 
				 (char *)&select_resp, sizeof(select_resp));
	if (nwritten != sizeof(select_resp)) {
		printf("Error writing select_resp\n");
		retval = -3;
		goto end;
	}

#define write_fd_mask(name, nfds) \
	do { \
		int __len = howmany(nfds, NFDBITS)*sizeof(fd_mask); \
		assert(name != NULL); \
		nwritten = write_n_bytes(fd, (char *)name, __len); \
		if (nwritten != __len) { \
			printf("Error writing select_req %s fd_mask\n", \
				#name); \
			retval = -2; \
			goto end; \
		} \
	} while (0)

	write_fd_mask(in, nfd);
	write_fd_mask(out, nfd);
	write_fd_mask(ex, nfd);
#undef write_fd_mask

end:
	if (in != NULL) {
		alpine_free(in);
	}
	if (out != NULL) {
		alpine_free(out);
	}
	if (ex != NULL) {
		alpine_free(ex);
	}
}

int
request_func_fork(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct fork_req fork_req;
	struct fork_resp fork_resp;

	assert(fd >= 0);
	assert(rtype == REQ_FORK);

	nread = read_n_bytes(fd, 
			     (char *)&fork_req + REQ_FIELD_SIZE,
			     sizeof(fork_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(fork_req) - REQ_FIELD_SIZE) {
		printf("Error reading fork_req\n");
		retval = -1;
		goto end;
	}

	fork_resp.resp_type = RESP_FORK;
	fork_resp.retval = 0; // TODO accounting fd for fork!
	fork_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, (char *)&fork_resp, sizeof(fork_resp));
	if (nwritten != sizeof(fork_resp)) {
		printf("Error writing fork_resp\n");
		retval = -2;
		goto end;
	}

end:
	return (retval);
}

int
request_func_dup(int fd, enum request_type rtype)
{
	int retval = 0, nread, nwritten;
	struct dup_req dup_req;
	struct dup_resp dup_resp;

	assert(fd >= 0);
	assert(rtype == REQ_DUP);

	nread = read_n_bytes(fd, 
			     (char *)&dup_req + REQ_FIELD_SIZE,
			     sizeof(dup_req) - REQ_FIELD_SIZE);
	if (nread != sizeof(dup_req) - REQ_FIELD_SIZE) {
		printf("Error reading dup_req\n");
		retval = -1;
		goto end;
	}

	dup_resp.resp_type = RESP_DUP;
	// TODO management of dup
	dup_resp.retval = alpine_dup(dup_req.old_fd);
	dup_resp.alpine_errno = alpine_internal_errno;

	nwritten = write_n_bytes(fd, (char *)&dup_resp, sizeof(dup_resp));
	if (nwritten != sizeof(dup_resp)) {
		printf("Error writing dup_resp\n");
		retval = -2;
		goto end;
	}

end:
	return (retval);
}
