/*   This is part of ALE4Net
 *   Application Level Environment for Networking
 *   
 *   req_resp: server/client message handling routines
 *
 *   Copyright 2004 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */   
/*
 * Neelkanth Natu (neelnatu@yahoo.com) is the author of this file.
 * As long as you keep this blurb in the file, feel free to do whatever.
 */ 
#include <sys/types.h>
#include <sys/socket.h>
#include <assert.h>

#include "req_resp.h"

const char *request_type_to_string[(1 + REQ_LAST)] = {
	"(n/a)",
	"socket",
	"close",
	"bind",
	"listen",
	"accept",
	"connect",
	"read",
	"write",
	"getsockname",
	"getpeername",
	"recv",
	"recvfrom",
	"recvmsg",
	"send",
	"sendto",
	"sendmsg",
	"shutdown",
	"setsockopt",
	"getsockopt",
	"fcntl",
	"select",
	"select_stop",
	"fork",
	"dup",
	"ioctl",
};
const char *response_type_to_string[1 + RESP_LAST] = {
	"(n/a)",
	"socket",
	"close",
	"bind",
	"listen",
	"accept",
	"connect",
	"read",
	"write",
	"getsockname",
	"getpeername",
	"recv",
	"recvfrom",
	"recvmsg",
	"send",
	"sendto",
	"sendmsg",
	"shutdown",
	"setsockopt",
	"getsockopt",
	"fcntl",
	"select",	/* For both select and select_stop requests */
	"fork",
	"dup",
	"ioctl"
};

/* 
 * XXX: We don't do ntohl anywhere because the client is supposed to be
 * running on the same machine as the alpine server. Even if we did do
 * ntoh and hton for the fields we know, we cannot do so for things
 * that we don't know about - like struct ifreq for SIOCSIFADDR.
 * So we don't bother.
 */

int
read_n_bytes(int fd, char *buf, int nbytes)
{
	int n, nread;

	assert(fd >= 0);
	assert(buf != NULL);
	assert(nbytes > 0);

	nread = 0;
	while (nbytes > 0) {
		n = native_read(fd, buf, nbytes);
		if ((n == -1) || (n == 0)) {
			/* EOF or error */
			break;
		}
		assert(n > 0);
		nread += n;
		nbytes -= n;
		buf += n;
	}

	return (nread);
}

int
write_n_bytes(int fd, const char *buf, int nbytes)
{
	int n, nwritten;

	assert(fd >= 0);
	assert(buf != NULL);
	assert(nbytes > 0);

	nwritten = 0;
	while (nbytes > 0) {
		n = native_write(fd, buf, nbytes);
		/* XXX: Catch SIGPIPE */
		if ((n == -1) || (n == 0)) {
			/* EOF or error */
			break;
		}
		assert(n > 0);
		nwritten += n;
		nbytes -= n;
		buf += n;
	}

	return (nwritten);
}

enum request_type
read_req_type(int fd)
{
	int nread;
	u_int32_t cmd;

	nread = read_n_bytes(fd, (char *)&cmd, sizeof(cmd));
	if (nread != sizeof(cmd)) {
		return (REQ_NONE);	/* Error */
	} 

	return (cmd);
}

enum response_type
read_resp_type(int fd)
{
	int nread;
	u_int32_t resp;

	nread = read_n_bytes(fd, (char *)&resp, sizeof(resp));
	if (nread != sizeof(resp)) {
		return (RESP_NONE);	/* Error */
	} 

	return (resp);
}

int
write_resp_type(int fd, enum response_type resp_type)
{
	int nwritten;

	assert((resp_type > RESP_NONE) && (resp_type <= RESP_LAST));

	nwritten = write_n_bytes(fd, (char *)&resp_type, sizeof(resp_type));
	if (nwritten != sizeof(resp_type)) {
		return (-1);
	}

	return (0);
}

int
write_req_type(int fd, enum request_type req_type)
{
	int nwritten;

	assert((req_type > REQ_NONE) && (req_type <= REQ_LAST));

	nwritten = write_n_bytes(fd, (char *)&req_type, sizeof(req_type));
	if (nwritten != sizeof(req_type)) {
		return (-1);
	}

	return (0);
}
